from loguru import logger

from client.schnell.workers import work
from common import mvc
from common.mvc.Models import Task, FailedTask, TaskPerformance
from common.networking.intercom import Intercom
from client.schnell import workers
from common.utils.io import getFilesInFileFolder
from os import getenv as env


class Client:
	jobs: dict = None
	intercom: Intercom = None

	def __init__(self):
		self.jobs = {}

	def mark_as_failed(self, session, message):
		item = session.query(Task).filter_by(id=message["task_id"]).first()
		if item is None:
			logger.error("Cannot mark task {} as failed as it does not exist in the DB.".format(message["task_id"]))
			return False

		if item.tries < 3:
			item.tries += 1
			session.add(item)
			logger.info("Incremented failure counter for task {}".format(item.id))
		else:
			failed = FailedTask(item.module_id, 3, str(message['error']))
			session.add(failed)
			session.delete(item)
			logger.warning("Deleted task {}.".format(item.id))

		session.commit()


	def mark_as_done(self, session, message):
		item = session.query(Task).filter_by(id=message["task_id"]).first()
		if item is None:
			logger.error("Cannot mark task {} as done as it does not exist in the DB.".format(message["task_id"]))
			return False

		perf = TaskPerformance(item.module_id, message['time'])
		session.add(perf)
		session.delete(item)
		session.commit()
		logger.warning("Marked task {} as done.".format(item.id))

	def serve(self, allowRemote):
		# Handles the client daemon
		intercom = Intercom("*" if allowRemote else "127.0.0.1", int(env("CLIENT_PORT")), False)
		session = mvc.getSession()

		logger.info("Listening for messages...")
		while True:
			message = intercom.listen()
			logger.info("Received message")
			if not type(message) is dict:
				logger.error("Received invalid message.")
				intercom.ko()
			elif "type" not in message:
				logger.error("Received corrupted message. Throwing away.")
				intercom.ko()
			else:
				# SHUT DOWN
				if message["type"] == "shutdown":
					logger.warning("==> | !!! Shutdown !!! | <==")
					intercom.ok()
					break

				# HEART BEAT
				elif message["type"] == "heartbeat":
					logger.debug("Received heartbeat, sending alive signal.")
					intercom.talk({"type": "alive"})

				# JOBS
				elif message["type"] == "jobs":
					if "jobs" in message:
						logger.info("Received jobs, adding to queue")
						logger.info("Jobs ID : " + str([k['id'] for k in message["jobs"]]))
						workers.pool.map_async(work, message["jobs"])
						intercom.ok()
					else:
						logger.error("Jobs order is corrupted, ignoring")
						intercom.ko()

				# FAIL NOTICE
				elif message["type"] == "failed":
					if "task_id" in message and "error" in message:
						self.mark_as_failed(session, message)
						intercom.ok()
					else:
						logger.error("Failed order is corrupted, ignoring")
						intercom.ko()

				# SUCCESS NOTICE
				elif message["type"] == "done":
					if "task_id" in message and "time" in message:
						self.mark_as_done(session, message)
						intercom.ok()
					else:
						logger.error("Done order is corrupted, ignoring")
						intercom.ko()

				# STATREP
				elif message["type"] == "statrep":
					files = getFilesInFileFolder()
					processPool = workers.nWorkers
					intercom.talk({"processPool": processPool, "files": files})

				# UNKNOWN MESSAGE
				else:
					logger.error("Unrecognised message. Ignoring.")
					intercom.ko()

		logger.info("Closing session")
		session.close()
