import multiprocessing
import time
from os import getenv as env

from loguru import logger

from common.mvc.Models import process
from common.networking.intercom import Intercom


def evaluateWithContext(argument: str, context: dict):
	if not isinstance(argument, str):
		return argument

	if argument.startswith("!") and argument[1:] in context:
		return context[argument[1:]]
	else:
		return argument


def work(item, context: dict = None):
	logger.add("run_client.log", format="{time} \t| {process.id} \t| {level}       | {message}", enqueue=True)
	context = context if context is not None else {}
	intercom = Intercom("127.0.0.1", env("CLIENT_PORT"), True)

	logger.info("Got item to work on !")

	# Checking item
	if item is None:
		logger.warning("This task does not exist anymore ! Skipping...")
	else:
		try:
			logger.debug("Getting ready to launch item...")
			chrono = time.time()
			logger.debug("Ready. Running item...")
			arguments = {k: evaluateWithContext(v, context) for k, v in item['arguments'].items()}
			resultFromProcess = process(item, arguments)
			logger.info("Item processed.")
			context = {**context, **resultFromProcess}

			# Sending job done notification
			logger.debug("Calculating time...")
			chrono = time.time() - chrono
			intercom.talk({"type": "done", "task_id": item["id"], "time": int(chrono)})
			intercom.listen()

			# Next chain
			if item.get('next_task') is not None:
				child = item['next_task']
				logger.info("Received instructions to work on specific task (child), running.")
				work(child, context)
				logger.info("Child processed.")
		except AssertionError as e:
			intercom = Intercom("127.0.0.1", env("CLIENT_PORT"), True)
			intercom.talk({"type": "failed", "task_id": item["id"], "error": str(e)})
			intercom.listen()
			work(item, context)
		except Exception as e:
			logger.critical("Unhandled error in worker code !")
			logger.error(e)
			intercom = Intercom("127.0.0.1", env("CLIENT_PORT"), True)
			intercom.talk({"type": "failed", "task_id": item["id"], "error": str(e)})
			intercom.listen()
		finally:
			intercom.socket.close()
		logger.success("Closing down task.")


class Schnell:
	nWorkers: int = 0
	pool: multiprocessing.Pool = None

	def __init__(self, nWorkers: int):
		self.nWorkers = nWorkers

	def start(self):
		logger.info("Starting pool with {0} workers...".format(self.nWorkers))
		self.pool = multiprocessing.get_context("spawn").Pool(processes=self.nWorkers)

	# Stops the schnell pool.
	def stop(self):
		logger.warning("Stopping schnell pool...")
		self.pool.terminate()
		logger.warning("Done.")
