from os import getenv as env

from loguru import logger
from sqlalchemy import Column, BigInteger, Text, Boolean, JSON, Integer, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from common.utils import starters

Base = declarative_base()


class Module(Base):
	__tablename__ = 'modules'

	# Fields
	id = Column(BigInteger, primary_key=True)
	name = Column(Text)
	usesSelenium = Column(Boolean)
	startArguments = Column(JSON)

	# Relationship
	tasks = relationship('Task', back_populates='module')

	def __init__(self, name, usesSelenium, startArguments):
		self.name = name
		self.usesSelenium = usesSelenium
		self.startArguments = startArguments

	def __repr__(self):
		return "<Module named {0}, does {1} use Selenium with start arguments {2}>" \
			.format(self.name, "" if self.usesSelenium else "not", self.startArguments)

	def to_dict(self):
		return {
			"id": self.id,
			"name": self.name,
			"usesSelenium": self.usesSelenium,
			"startArguments": self.startArguments
		}


class Task(Base):
	__tablename__ = 'tasks'

	# Fields
	id = Column(BigInteger, primary_key=True)
	module_id = Column(BigInteger, ForeignKey(Module.id))
	arguments = Column(JSON)
	tries = Column(Integer)

	# Relationships
	module = relationship('Module', uselist=False)

	def __init__(self, module_id, arguments, tries, next_task=None):
		self.module_id = module_id
		self.arguments = arguments
		self.tries = tries
		self.next_task = next_task

	def time(self):  # Time estimations of modules in seconds
		if self.module.name == "Swiss Impex":
			return 180
		elif self.module.name == "FAO Genève":
			return 20
		elif self.module.name == "XLSX Parser":
			return 3
		else:
			return 0

	def __repr__(self):
		task = "<Task for module {0}, with arguments {1}. {2} tries.>" \
			.format(self.module_id, self.arguments, self.tries)
		if self.next_task is not None:
			return task + ". Next : " + self.next_task.__repr__()

	def to_dict(self):
		return {
			"id":           self.id,
			"module_id":    self.module_id,
			"module":       self.module.to_dict(),
			"arguments":    self.arguments,
			"tries":        self.tries,
			"next_task":    self.next_task
		}


def process(item: dict, arguments: dict) -> dict:
	logger.debug("Seeking correct module...")
	mName = item["module"]["name"]
	if mName == "Swiss Impex":
		return starters.startSwissImpex(
			arguments["merchandiseId"], arguments["year"], arguments["month"],
			int(env("HTTPHANDLER_WEBSERVER_PORT"))
		)

	elif mName == "FAO Genève":
		starters.startFAOGeneve(arguments["pages"])
		return {}

	elif mName == "XLSX Parser":
		return starters.startXSLXParser(arguments["file"], arguments["instructions"], arguments.get('index') or None,
		                                [arguments.get('from') or None, arguments.get('to') or None])
	elif mName == "File Parser":
		return starters.startFileParser(arguments["file"], arguments["instructions"])

	elif mName == "Regex Parser":
		return starters.startRegexParser(arguments["input"], arguments["regex"])

	else:
		# TODO In fact, this will be done in the bachelor part. For now, we are only using static modules.
		# ADD HERE FOR CUSTOM MODULES
		logger.error("Unrecognized module.")


# JIT Relationships
Task.next_task = Column(BigInteger, ForeignKey(Task.id))


class FailedTask(Base):
	__tablename__ = 'failed_tasks'

	# Fields
	id = Column(BigInteger, primary_key=True)
	module_id = Column(BigInteger, ForeignKey(Module.id))
	tries = Column(Integer)
	message = Column(Text)

	# Relationships
	module = relationship('Module', uselist=False)

	def __init__(self, module_id, tries, message):
		self.module_id = module_id
		self.tries = tries
		self.message = message


class TaskPerformance(Base):
	__tablename__ = 'task_performance'

	# Fields
	id = Column(BigInteger, primary_key=True)
	module_id = Column(BigInteger, ForeignKey(Module.id))
	time = Column(Integer)

	# Relationships
	module = relationship('Module', uselist=False)

	def __init__(self, module_id, performance):
		self.module_id = module_id
		self.time = performance

