import copy
import itertools
from collections import OrderedDict

from sqlalchemy.orm import Session
from zmq import ZMQError

from client.schnell.workers import evaluateWithContext
from common.networking.intercom import Intercom
from common import mvc
import yaml
from loguru import logger
from common.mvc.Models import Module, Task, process
from os import getenv as env


class YAMLParser:
	file = None

	def __init__(self, path):
		self.file = yaml.load(open(path), Loader=yaml.FullLoader)

	# Converts clients inside the YAML space into real full blown IPs
	def evaluateClients(self, isRemote: bool, shouldPing: bool = True) -> list:
		if "clients" not in self.file and isRemote:
			logger.error("No clients defined in playbook despite remote mode enabled. Stopping.")
			exit(3)

		clientsInput = self.file.get("clients")
		output = []
		if not isRemote:
			clientsInput = ["127.0.0.1"]
		if not shouldPing:
			return self.file.get("clients")

		for client in clientsInput:
			logger.info("Checking heartbeat for {0}...".format(client))
			try:
				intercom = Intercom(client, int(env("CLIENT_PORT")), True)
				intercom.talk({"type": "heartbeat", "payload": ""})
				message = intercom.listen(750)
				if "type" not in message:
					logger.error("Received corrupted message from client {0}.".format(client))
					raise RuntimeError
				output.append(client)
			except (ZMQError, RuntimeError):
				logger.error("Cannot talk to {0} !".format(client))

		return output

	# Variables beginning and ending with a '$' can be
	def parseVariable(self, key: str):
		key = str(key)
		if key.startswith("$") and key.endswith("$"):
			variables = self.file.get("variables")[0]
			key = key[1:-1]
			if key in variables:
				return variables[key]
			else:
				logger.error("{} does not exist within playbook !".format(key))
				exit(4)
		else:
			return key

	# Creates the jobs, sends them to their owner and puts them into the DB.
	def evaluateJobs(self, isDry):
		logger.info("Starting playbook...")

		# SECOND STEP : READING CHAPTERS
		chapters = self.file.get("chapters", [])
		logger.info("Detected {0} chapter(s) to play.".format(len(chapters)))
		if len(chapters) == 0:
			logger.error("No chapters to play. Stopping.")
			return

		# Generating commands for further processing
		actions = []
		context = {}
		clientLock = False
		for chapter, args in chapters.items():
			# Immediately checking for pass keyword
			if "pass" in args and args.get("pass"):
				logger.info("Chapter {} has pass keyword, skipping".format(chapter))
				continue
			elif "server" in args and args.get("server"):
				if clientLock:
					logger.error("Server tasks cannot be run anymore because some client tasks have been generated.")
					exit(20)
				else:
					logger.warning("Chapter {} has server keyword, running server tasks now.".format(chapter))
					context, clientModule, clientModuleName = self.__iterateOnServerChapter__(chapter, args, {})
					logger.warning("Server tasks are now done.")
					# Manually calling client tasks. Required since all of then tasks are one single chapter
					actions += self.__iterateOnChapter__(clientModuleName, clientModule, context, isDry)
			else:
				clientLock = True
				actions += self.__iterateOnChapter__(chapter, args, {}, isDry)

		return chapters, actions

	# Runs immediate actions and returns the generated infos
	def __iterateOnServerChapter__(self, chapter, args: dict, context: dict) -> (dict, dict):
		# Trying to find module in DB
		session = mvc.getSession()

		module: Module = session.query(Module).filter_by(name=chapter).first()
		if module is None:
			logger.error("This module does not exist within the database.")
			exit(5)

		output = process({"module": module.to_dict()}, {k: evaluateWithContext(arg, context) for k, arg in args.items()})
		context = {**output, **context}
		nextModuleName = list(args["then"].keys())[0] if "then" in args else None

		if "then" in args:
			if "server" in args["then"][nextModuleName] and args["then"][nextModuleName]["server"]:
				logger.info("Server item has then module, running.")

				lasts, nextModule, name = self.__iterateOnServerChapter__(nextModuleName, args["then"][nextModuleName], context)
				return {**output, **lasts}, nextModule, name

		if "then" not in args:
			logger.error("Invalid YAML Format : Missing client tasks.")
			exit(60)

		if "server" not in args["then"][nextModuleName]:
			logger.info("Last server item detected, stopping")
			return output, args["then"][nextModuleName], nextModuleName

	# Iterates on one chapter
	def __iterateOnChapter__(self, chapter, args, context, isDry):
		logger.info("==========================================================")
		logger.info("Searching for module named as chapter " + chapter + "...")

		# Trying to find module in DB
		session = mvc.getSession()
		module = session.query(Module).filter_by(name=chapter).first()
		if module is None:
			logger.error("This module does not exist within the database.")
			exit(5)
		logger.success("Found module in database.")

		logger.info("Parsing chapter " + chapter + "...")
		expansions = OrderedDict()  # Contains all generated expansions items. The expansion is NOT done here !
		meta = OrderedDict()  # Contains all non expanded key with their values

		# Parsing arguments
		for argName, argOptions in args.items():
			if argName == "then":
				logger.info("Found then item, marking as to do later")
				meta[argName] = next(iter(argOptions.keys()))
			elif isinstance(argOptions, dict):
				logger.debug("Processing complex argument " + argName + "...")
				expansions[argName] = self.__handleComplexType__(argName, argOptions)
			else:
				logger.info("Simple argument {0} considered as a meta argument.".format(argName))
				meta[argName] = self.parseVariable(argOptions)

		logger.info("Finished reading file, now generating objects for processing")

		tasks = [OrderedDict(zip(list(expansions.keys()), task)) for task in
		         list(itertools.product(*expansions.values()))]

		logger.info("Generated " + str(len(tasks)) + " tasks.")

		# Meta argument injections
		if len(meta) > 0:
			logger.info("Now trying to add meta arguments...")
			for argName, argValues in meta.items():
				logger.info("Adding " + argName + "...")
				for item in range(len(tasks)):
					tasks[item][argName] = argValues

		# Server injection tasks generation
		if len(context) > 0:
			logger.info("Injecting server variables into client tasks.")
			realTasks = []
			for task in tasks:
				newBatch = []
				results = [(k, v) for k, v in list(task.items()) if str(v).startswith("!")]
				if len(results) > 1:
					logger.error("Server arguments are supported for only one key for now.")

				argKey, argVal = results[0]
				evaluatedArgument = evaluateWithContext(argVal, context)

				for eA in evaluatedArgument:
					newTask = copy.deepcopy(task)
					newTask[argKey] = eA
					newBatch.append(newTask)

				realTasks += newBatch
			tasks = realTasks

		# Following immediately the then items. Allows for proper task workflow guidance
		thenItems = None
		if "then" in meta:
			logger.info("Doing 'then' items now. First checking if then module is present in DB...")
			thenModule = session.query(Module).filter_by(name=meta["then"]).first()
			if thenModule is not None:
				thenItems = self.__handleThenType__(args["then"], isDry, thenModule, len(tasks), session)
			else:
				logger.error("Then module is not found in DB.")
				exit(15)

		# Sending into DB
		logger.info("Now converting into tasks...")
		if len(tasks) > 10000:
			logger.warning("This is an extremely heavy operation. This WILL take some time. Please wait.")
		output = list()
		for k in range(len(tasks)):
			task = Task(module.id, tasks[k], 0, None if thenItems is None else thenItems[k].id)
			output.append(task)
			if not isDry:
				session.add(task)
		if not isDry:
			session.commit()

		logger.success("Finished parsing chapter.")

		if isDry:
			realOutput = [task.__repr__() for task in output]
		else:
			# Small hack to say no to SQLAlchemy...
			for k in range(len(output)):
				output[k].next_task = thenItems[k].to_dict() if thenItems is not None else None

			realOutput = [task.to_dict() for task in output]

		session.close()

		return realOutput

	# Handles a complex option in the YAML file
	def __handleComplexType__(self, argName: str, argOptions: dict):
		# Range option
		if "start" in argOptions and "end" in argOptions:
			logger.debug("Detected arguments for range generation, creating...")

			start, end = int(self.parseVariable(argOptions["start"])), int(self.parseVariable(argOptions["end"]))
			gen = range(start, end)

			return list(gen)
		else:
			return [argOptions]

	# Handles "then" instructions
	def __handleThenType__(self, argOptions: OrderedDict, isDry: bool, module: Module, context: int, session : Session) -> list:
		if len(argOptions) > 1:
			logger.error("Multiple then items are not yet supported.")
			# Non supported (MySQL current format), INOP
		else:
			for thenModuleName, thenModuleArguments in argOptions.items():
				if type(thenModuleArguments) == dict and "start" in thenModuleArguments and "end" in thenModuleArguments:
					logger.error("Then tasks do not accept complex arguments that expand.")
					exit(10)

			output = []
			for k in range(context):
				t = Task(module.id, argOptions[next(iter(argOptions.keys()))], 0, None)
				output.append(t)
				if not isDry:
					session.add(t)
			if not isDry:
				session.commit()

			return output



