from os import getenv as env
from threading import Thread

import zmq
from flask import render_template, request, abort
from loguru import logger

from common.networking.intercom import Intercom
from common.utils.LoguruHandler import FlaskToLoguruHandler
from common import mvc
from server.mvc import app
from common.mvc.Models import Module, Task, FailedTask, TaskPerformance

session = mvc.getSession()
nodes: dict = {}


# Sets up the app
def setup():
	fao = Module("FAO Genève", False, {})

	swissimpex = Module("Swiss Impex", True,
	                    {"merchId": "text", "month": "number", "year": "number", "webServerPort": "number"})

	xslxParser = Module("XLSX Parser", False, {})
	fileParser = Module("File Parser", False, {"file": "string"})
	regexParser = Module("Regex Parser", False, {"file": "string"})
	# FOR CUSTOM MODULES : ADD YOUR MODULE HERE
	session.add(fao)
	session.add(swissimpex)
	session.add(xslxParser)
	session.add(fileParser)
	session.add(regexParser)
	session.commit()


class WebServer:
	# Views - Used for sending the views to the client
	@staticmethod
	@app.route("/")
	def showWelcomeScreen():
		tasks = session.query(Task).all()
		nodeInfos = {}

		# Filling up nodeInfos
		for key, node in nodes.items():
			intercom = Intercom(node, int(env("CLIENT_PORT")), True)
			try:
				intercom.talk({"type": "statrep"})
				nodeInfos[node] = intercom.listen(500)
			except zmq.Again:
				logger.critical("Cannot contact host {} ! Will be removed from clients.".format(node))
			finally:
				intercom.socket.close()

		seconds = sum(map(lambda task: task.time(), tasks))
		totalWorkers = sum(map(lambda n: n["processPool"], nodeInfos.values()))
		seconds /= totalWorkers if totalWorkers > 0 else 1
		# https://stackoverflow.com/questions/775049/how-do-i-convert-seconds-to-hours-minutes-and-seconds
		m, s = divmod(seconds, 60)
		h, m = divmod(m, 60)

		# JIT Setup for webserver
		if session.query(Module).count() == 0:
			setup()

		return render_template('index.html', nodes=nodes, modules=session.query(Module).count(),
		                       tasks=len(tasks), estimatedTime='{:.2f} heures, {:.2f} minutes et {:.2f} secondes'
		                       .format(h, m, s), nodeInfos=nodeInfos, clientPort=str(env("HTTPHANDLER_WEBSERVER_PORT")))

	@staticmethod
	@app.route("/modules", methods=["GET"])
	def showModules():
		return render_template('modules.html', modules=session.query(Module).all())

	# Misc - Used for miscellaneous reasons
	@staticmethod
	@app.route("/shutdown", methods=["GET", "POST"])
	def shutdown():
		if request.method == "POST":
			session.close()
			func = request.environ.get('werkzeug.server.shutdown')
			func()
			return "Le système a correctement été arrêté. Au revoir."
		else:
			return render_template("shutdown.html")

	# noinspection DuplicatedCode
	@staticmethod
	@app.route("/task/<task>/failed", methods=["POST"])
	def mark_as_failed(task):
		item = session.query(Task).filter_by(id=task).first()
		if item is None:
			abort(404)
		failed = FailedTask(item.module_id, 3, str(request.form['error'] if 'error' in request.form else ''))
		session.add(failed)
		session.delete(item)
		session.commit()

		return "OK"

	# noinspection DuplicatedCode
	@staticmethod
	@app.route("/job/<task>/done", methods=["POST"])
	def mark_as_done(task):
		item = session.query(Task).filter_by(id=task).first()
		if item is None:
			abort(404)
		perf = TaskPerformance(item.module_id, request.form['time'])
		session.add(perf)
		session.delete(item)
		session.commit()

	# Startup method

	@staticmethod
	def start(numberOfNodes):
		def realRun():
			global nodes
			nodes = {i: numberOfNodes[i] for i in range(len(numberOfNodes))}
			app.logger.addHandler(FlaskToLoguruHandler)
			app.run(host="0.0.0.0", port=8080)

		Thread(target=realRun).start()
