import argparse

from client import Client
from common.networking.intercom import Intercom
from common import mvc
import requests

from dotenv import load_dotenv, find_dotenv
from loguru import logger

from common.utils.PerfObserver import PerformanceObserver
from server.mvc.WebServer import WebServer
from client.schnell import workers
from common.mvc.Models import Task
from common.utils.YAMLParser import YAMLParser
from os import getenv as env
from server.dispatch.shotgun import Shotgun

# Environment awareness activation
load_dotenv(find_dotenv())

# Option parsing and checking
parser = argparse.ArgumentParser(description='Flexible and scalable document parser')
group = parser.add_mutually_exclusive_group()
group.add_argument("--server", help="Considers this instance as a server", default=False, action='store_true')
group.add_argument("--client", help="Considers this instance as a client. On by default.", default=True,
                   action='store_true')

parser.add_argument("--wipe-tasks", help="Truncates the tasks table in the database. This option is considered quite "
                                         "dangerous as it WILL break any running worker. Be careful !", default=False,
                    action='store_true')
parser.add_argument("--dry", help="Does not actually send any task anywhere. Just shows the jobs produced by the YAML "
                                  "file.", default=False, action='store_true')
parser.add_argument("--remote", help="Allows remote jobs to be sent to this instance. Only used when working in "
                                     "client mode.", default=False, action='store_true')
parser.add_argument("--playbook", help="The playbook to run. Only used when working in server mode.")
parser.add_argument("--perf-files", help="Monitor processed files in file locker for performance analysis",
                    default=False, action='store_true')
parser.add_argument("--perf-db", help="Monitor tasks in central database for performance analysis",
                    default=False, action='store_true')
parser.add_argument("--perf-timing", help="Timing between each performance analysis measurement. Default : 1",
                    default=1)

args = parser.parse_args()

# Identity awareness evaluation
isServer = args.server
cleanSlate = args.wipe_tasks
isDry = args.dry
allowRemote = args.remote
playbook = args.playbook
shouldMonitorFiles = args.perf_files
shouldMonitorDB = args.perf_db
shouldMonitor = shouldMonitorFiles or shouldMonitorDB
monitorTiming = args.perf_timing


# Executes a playbook. SERVER ONLY
def runPlaybook(p: YAMLParser, clts: list):
	c, t = p.evaluateJobs(isDry)
	logger.info("Created {0} jobs.".format(len(t)))

	if isDry:
		logger.warning("DRY MODE - No job will be dispatched or stored into the database, but the logs will tell you "
		               "how things will play out.")
		logger.info("Jobs found : \n{0}".format("\n".join(t)))

	shotgun = Shotgun(clts)
	r = shotgun.divideJobs(t)
	logger.info("Dispatching {0} jobs to {1} clients.".format(len(t), len(r)))
	if isDry:
		logger.info("Job repartition : \n{0}".format(r))

	return c, t, r


# Starting program
if __name__ == '__main__':
	if not isServer and shouldMonitor:
		logger.error("You cannot monitor performance with a Inari client.")
		exit(300)

	logger.debug("Connecting to DB...")
	session = mvc.getSession()

	if cleanSlate:
		logger.warning("All tasks will be wiped from the DB. This action cannot be undone without backup.")
		session.query(Task).delete()
		session.commit()
		logger.warning("All tasks have been deleted.")

	logger.info("Inari - Théo Pirkl")
	if not allowRemote:
		logger.warning("You have forbidden outside instances to connect to this instance. \nThat's fine, but you will "
		               "need to start another instance at the same time on this machine with the other mode in order "
		               "for everything to work.")
	if isServer:
		logger.add("run_server.log", format="{time} \t| {process.id} \t| {level}       | {message}", enqueue=True)
		logger.info("Running in server mode. All clients will be controlled from this server.")

		logger.info("=====================================")
		if playbook:
			playbook = YAMLParser(playbook)
			clients = playbook.evaluateClients(allowRemote)
			logger.info("Starting management web interface...")
			WebServer.start(clients)
			logger.info("Press enter to read YAML playbook.")
			input()
			chapters, tasks, repartition = runPlaybook(playbook, clients)

			if len(clients) == 0:
				logger.critical("No client detected. Will continue pointlessly.")

			input("Press enter to send {} tasks ({} chapters).".format(len(tasks), len(chapters)))
			for client, jobs in repartition.items():
				clientIntercom = Intercom(client, int(env("CLIENT_PORT")), True)
				clientIntercom.talk({"type": "jobs", "jobs": jobs})
				clientIntercom.listen()

			# Launching monitoring if needed
			po = PerformanceObserver(clients, monitorTiming)
			if shouldMonitor:
				po.start(shouldMonitorDB, shouldMonitorFiles)

			input("Press enter to shut down program. This will shut down the management web server.\n"
			      "It's a better practice to wait for the jobs to fully finish before stopping this server.")
			po.stop()
			k = input("Press enter to shut down all workers. Type anything else to shut down only the server.")
			if k == "":
				Intercom.broadcastToClients(clients, {"type": "shutdown"})
		else:
			logger.error("Please enter a playbook. Press enter to shutdown.")
			input()

		logger.info("Shutting down...")
		requests.post("http://127.0.0.1:8080/shutdown")

	else:
		logger.add("run_client.log", format="{time} \t| {process.id} \t| {level}       | {message}", enqueue=True)
		# Specific imports
		from client.spiders.seleniumspiders import HTTPHandler

		logger.info("Running in client mode. All order received by any server will be processed.")
		logger.info("Starting local management web server...")
		HTTPHandler.start(int(env("HTTPHANDLER_WEBSERVER_PORT")))

		# Starting workers. Jobs can be executed below this line !!
		logger.info("Starting workers...")
		workers.start()

		logger.info("Done. Ready to process orders...")
		client = Client()
		client.serve(allowRemote)

		logger.info("Received shutdown signal. Stopping.")
		logger.info("Stopping workers...")
		workers.stop()
		# Stopped workers. Jobs won't be executed below this line.

		logger.info("Closing HTTP Handler...")
		requests.get("http://127.0.0.1:{0}/shutdown".format(int(env("HTTPHANDLER_WEBSERVER_PORT"))))
		logger.info("Closing database session...")

	session.close()
	# Everything has been stopped.
	logger.success("All services have been shut down. Good night.")
